
{*******************************************************}
{                                                       }
{       RichViewActions                                 }
{       Ancestor for all RichViewActions forms.         }
{                                                       }
{       Copyright (c) 2002-2004, Sergey Tkachenko       }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

unit BaseRVFrm;

interface

{$I RichViewActions.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, StdCtrls, ExtCtrls,
  Dialogs, Buttons, RVScroll, RVXPTheme,
  {$IFDEF USERVKSDEVTE}
  te_controls, CRVPP, RVPP, RichView, te_richview,
  {$ENDIF}
  RVOfficeRadioBtn, ComCtrls;

type
  TfrmRVBase = class(TForm)
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    procedure btnHelpClick(Sender: TObject);
  protected
    procedure SetUseXPThemes(const Value: Boolean); virtual;
    procedure LocalizeSpecialControls;
    function FindButtonWithModalResult(ModalResult: TModalResult): TButton;
    procedure AddHelpButton;
    function GetOkButton: TButton; dynamic;
    function GetCancelButton: TButton;
    {$IFDEF USERVKSDEVTE}
    function IsTeFormAllowed: Boolean; dynamic;
    procedure OnCreateThemedControl(OldControl, NewControl: TControl); dynamic;
    function IsThemeAllowedFor(Component: TComponent): Boolean; dynamic;
    procedure MoveChildren(OldControl, NewControl: TWinControl);
    procedure ApplyKSDevThemeEngine;
    {$ENDIF}
  public
    { Public declarations }
    {$IFDEF USERVKSDEVTE}
    tef: TTeForm;
    {$ENDIF}
    procedure Localize; dynamic;
    function GetRadioButtonChecked(rb: TControl): Boolean;
    procedure SetRadioButtonChecked(rb: TControl; Checked: Boolean);
    function GetCheckBoxChecked(cb: TControl): Boolean;
    procedure SetCheckBoxChecked(cb: TControl; Checked: Boolean);
    function GetCheckBoxState(cb: TControl): TCheckBoxState;
    procedure SetCheckBoxState(cb: TControl; State: TCheckBoxState);
    procedure SetFormCaption(const Caption: String);
    procedure SetControlCaption(Control: TControl; const Caption: String);
    function GetEditText(Edit: TControl): String;
    procedure SetEditSelText(Edit: TControl; const s: String);    
    function GetEditFont(Edit: TControl): TFont;
    function GetEditSelStart(Edit: TControl): Integer;
    function GetEditSelLength(Edit: TControl): Integer;
    procedure SetEditSelStart(Edit: TControl; Value: Integer);
    procedure SetEditSelLength(Edit: TControl; Value: Integer);
    function GetXBoxItems(Box: TControl): TStrings;
    function GetXBoxItemIndex(Box: TControl): Integer;
    function GetPageControlActivePageIndex(pc: TControl): Integer;
    function GetPageControlActivePage(pc: TControl): TWinControl;
    procedure SetPageControlActivePage(pc: TControl; Page: TControl);
    procedure SetXBoxItemIndex(Box: TControl; ItemIndex: Integer);
    procedure SetListBoxSorted(Box: TControl);
    procedure HideTabSheet(ts: TControl);
    function IsTabSheetVisible(ts: TControl): Boolean;
    property UseXPThemes: Boolean write SetUseXPThemes;
  end;

implementation
uses RichViewActions, RVColorCombo, RVALocalize, RVColorGrid, ColorRVFrm, RVUni;

{$R *.dfm}

function _GetWideString(const s: String): WideString;
begin
  Result := RVU_RawUnicodeToWideString(RVU_AnsiToUnicode(
    RVU_Charset2CodePage(RVA_GetCharset), s));
end;

{
function _GetString(const s: WideString): String;
begin
  Result := RVU_UnicodeToAnsi(RVU_Charset2CodePage(RVA_GetCharset),
    RVU_GetRawUnicode(s));
end;
}

{ TfrmRVBase }

function GetRevBiDiModeAlignment(Alignment: TAlignment): TAlignment;
begin
  case Alignment of
    taLeftJustify:  Result := taRightJustify;
    taRightJustify: Result := taLeftJustify;
    else Result := Alignment;
  end;
end;

procedure TfrmRVBase.SetUseXPThemes(const Value: Boolean);
var i: Integer;
begin
  for i := 0 to ComponentCount -1 do
    if Components[i] is TRVScroller then
      TRVScroller(Components[i]).UseXPThemes := Value
    else if Components[i] is TRVOfficeRadioButton then
      TRVOfficeRadioButton(Components[i]).UseXPThemes := Value
    else if Components[i] is TRVOfficeRadioGroup then
      TRVOfficeRadioGroup(Components[i]).UseXPThemes := Value
    else if Components[i] is TRVColorCombo then
      TRVColorCombo(Components[i]).UseXPThemes := Value
    else if Components[i] is TRVColorGrid then
      TRVColorGrid(Components[i]).Flat := Value and
       Assigned(RV_IsAppThemed) and RV_IsAppThemed and RV_IsThemeActive;
end;

procedure TfrmRVBase.FormCreate(Sender: TObject);
var i: Integer;
    Help: String;
begin
  Font.Charset := RVA_GetCharset;
  if Screen.Fonts.IndexOf(RVA_DialogFontName)>=0 then
    Font.Name := RVA_DialogFontName;
  UseXPThemes := RVA_UseXPThemes;
  if RVA_GetCharset in [ARABIC_CHARSET, HEBREW_CHARSET] then begin
    BiDiMode := bdRightToLeft;
    FlipChildren(True);
    if not SysLocale.MiddleEast then
      for i := 0 to ComponentCount-1 do
        if Components[i] is TCustomLabel then
          TLabel(Components[i]).Alignment := GetRevBiDiModeAlignment(TLabel(Components[i]).Alignment)
        else if Components[i] is TCustomCheckBox then
          TCheckBox(Components[i]).Alignment := GetRevBiDiModeAlignment(TCheckBox(Components[i]).Alignment)
        else if Components[i] is TRadioButton then
          TRadioButton(Components[i]).Alignment := GetRevBiDiModeAlignment(TRadioButton(Components[i]).Alignment)
  end;
  Localize;
  Help := RVA_GetHelpFile;
  if Help<>'' then begin
    HelpFile := ExtractFilePath(Application.ExeName)+Help;
    AddHelpButton;
  end;
  {$IFDEF USERVKSDEVTE}
  ApplyKSDevThemeEngine;
  {$ENDIF}
end;

procedure TfrmRVBase.Localize;
begin
  LocalizeSpecialControls;
end;

procedure TfrmRVBase.LocalizeSpecialControls;
var i: Integer;
    ColorPresent: Boolean;
begin
  ColorPresent := False;
  for i := 0 to ComponentCount-1 do
    if Components[i] is TRVColorCombo then begin
      TRVColorCombo(Components[i]).TransparentCaption := RVA_GetS(rvam_cl_Transparent);
      TRVColorCombo(Components[i]).AutoCaption := RVA_GetS(rvam_cl_Auto);
      ColorPresent := True;
      end
    else if Components[i] is TRVColorGrid then
      ColorPresent := True;
  if ColorPresent then
    for i := 0 to ColorCount-1 do
      Colors[i].Name := PChar(RVA_GetPC(TRVAMessageID(ord(rvam_cl_Black)+i)));
end;

function TfrmRVBase.FindButtonWithModalResult(
  ModalResult: TModalResult): TButton;
var i: Integer;
begin
  for i := 0 to ComponentCount-1 do
    if (Components[i] is TButton) and
       (TButton(Components[i]).ModalResult=ModalResult) then begin
      Result := TButton(Components[i]);
      exit;
    end;
  Result := nil;
end;

procedure TfrmRVBase.AddHelpButton;
var btnOk, btnCancel, btnHelp: TButton;
begin
  btnOk := GetOkButton;
  if btnOk=nil then
    exit;
  btnCancel := GetCancelButton;
  if btnCancel=nil then
    exit;
  btnHelp := TButton.Create(Self);
  btnHelp.Caption := RVA_GetS(rvam_btn_Help);
  with btnCancel do
    btnHelp.SetBounds(Left, Top, Width, Height);
  with btnOk do
    btnCancel.SetBounds(Left, Top, Width, Height);
  btnOk.Left := btnCancel.Left - (btnHelp.Left-btnCancel.Left);
  btnOk.Top := btnCancel.Top - (btnHelp.Top-btnCancel.Top);
  btnHelp.Parent := Self;
  btnHelp.TabOrder := btnCancel.TabOrder+1;
  btnHelp.OnClick := btnHelpClick;
end;

function TfrmRVBase.GetOkButton: TButton;
begin
  Result := FindButtonWithModalResult(mrOk);
end;

function TfrmRVBase.GetCancelButton: TButton;
begin
  Result := FindButtonWithModalResult(mrCancel);
end;

procedure TfrmRVBase.btnHelpClick(Sender: TObject);
begin
  Application.HelpContext(HelpContext);
end;

function TfrmRVBase.GetRadioButtonChecked(rb: TControl): Boolean;
begin
  {$IFDEF USERVKSDEVTE}
  Result := (rb as TTeRadioButton).Checked;
  {$ELSE}
  Result := (rb as TRadioButton).Checked;
  {$ENDIF}
end;

procedure TfrmRVBase.SetRadioButtonChecked(rb: TControl; Checked: Boolean);
begin
  {$IFDEF USERVKSDEVTE}
  (rb as TTeRadioButton).Checked := Checked;
  {$ELSE}
  (rb as TRadioButton).Checked := Checked;;
  {$ENDIF}
end;

function TfrmRVBase.GetCheckBoxChecked(cb: TControl): Boolean;
begin
  {$IFDEF USERVKSDEVTE}
  Result := (cb as TTeCheckBox).Checked;
  {$ELSE}
  Result := (cb as TCheckBox).Checked;
  {$ENDIF}
end;

procedure TfrmRVBase.SetCheckBoxChecked(cb: TControl; Checked: Boolean);
begin
  {$IFDEF USERVKSDEVTE}
  (cb as TTeCheckBox).Checked := Checked;
  {$ELSE}
  (cb as TCheckBox).Checked := Checked;;
  {$ENDIF}
end;

function TfrmRVBase.GetCheckBoxState(cb: TControl): TCheckBoxState;
begin
  {$IFDEF USERVKSDEVTE}
  Result := TCheckBoxState((cb as TTeCheckBox).State);
  {$ELSE}
  Result := (cb as TCheckBox).State;
  {$ENDIF}
end;

procedure TfrmRVBase.SetCheckBoxState(cb: TControl; State: TCheckBoxState);
begin
  {$IFDEF USERVKSDEVTE}
  (cb as TTeCheckBox).State := TTeCheckBoxState(State);
  {$ELSE}
  (cb as TCheckBox).State := State;
  {$ENDIF}
end;

function TfrmRVBase.GetXBoxItems(Box: TControl): TStrings;
begin
  {$IFDEF USERVKSDEVTE}
  if Box is TTeComboBox then
    Result := (Box as TTeComboBox).Items
  else if Box is TTeListBox then
    Result := (Box as TTeListBox).Items
  else
    Result := (Box as TTeRadioGroup).Items
  {$ELSE}
  if Box is TComboBox then
    Result := (Box as TComboBox).Items
  else if Box is TListBox then
    Result := (Box as TListBox).Items
  else
    Result := (Box as TRadioGroup).Items
  {$ENDIF}
end;

function TfrmRVBase.GetXBoxItemIndex(Box: TControl): Integer;
begin
  {$IFDEF USERVKSDEVTE}
  if Box is TTeComboBox then
    Result := (Box as TTeComboBox).ItemIndex
  else if Box is TTeListBox then
    Result := (Box as TTeListBox).ItemIndex
  else
    Result := (Box as TTeRadioGroup).ItemIndex  
  {$ELSE}
  if Box is TComboBox then
    Result := (Box as TComboBox).ItemIndex
  else if Box is TListBox then
    Result := (Box as TListBox).ItemIndex
  else
    Result := (Box as TRadioGroup).ItemIndex  
  {$ENDIF}
end;

procedure TfrmRVBase.SetXBoxItemIndex(Box: TControl;
  ItemIndex: Integer);
begin
  {$IFDEF USERVKSDEVTE}
  if Box is TTeComboBox then
    (Box as TTeComboBox).ItemIndex := ItemIndex
  else if Box is TTeListBox then
    (Box as TTeListBox).ItemIndex := ItemIndex
  else
    (Box as TTeRadioGroup).ItemIndex := ItemIndex;
  {$ELSE}
  if Box is TComboBox then
    (Box as TComboBox).ItemIndex := ItemIndex
  else if Box is TListBox then
    (Box as TListBox).ItemIndex := ItemIndex
  else
    (Box as TRadioGroup).ItemIndex := ItemIndex;
  {$ENDIF}
end;

procedure TfrmRVBase.SetListBoxSorted(Box: TControl);
begin
  {$IFDEF USERVKSDEVTE}
  (Box as TTeListBox).Sorted := True;
  {$ELSE}
  (Box as TListBox).Sorted := True;
  {$ENDIF}
end;

function TfrmRVBase.GetPageControlActivePageIndex(pc: TControl): Integer;
begin
  {$IFDEF USERVKSDEVTE}
  Result := (pc as TTePageControl).ActivePage.PageIndex;
  {$ELSE}
  Result := (pc as TPageControl).ActivePage.PageIndex;
  {$ENDIF}
end;

function TfrmRVBase.GetPageControlActivePage(pc: TControl): TWinControl;
begin
  {$IFDEF USERVKSDEVTE}
  Result := (pc as TTePageControl).ActivePage;
  {$ELSE}
  Result := (pc as TPageControl).ActivePage;
  {$ENDIF}
end;

procedure TfrmRVBase.SetPageControlActivePage(pc: TControl; Page: TControl);
begin
  {$IFDEF USERVKSDEVTE}
  (pc as TTePageControl).ActivePage := TTeTabSheet(Page);
  {$ELSE}
  (pc as TPageControl).ActivePage := TTabSheet(Page);;
  {$ENDIF}
end;

procedure TfrmRVBase.HideTabSheet(ts: TControl);
begin
  {$IFDEF USERVKSDEVTE}
  (ts as TTeTabSheet).PageVisible := False;
  {$ELSE}
  (ts as TTabSheet).TabVisible := False;
  {$ENDIF}
end;

function TfrmRVBase.IsTabSheetVisible(ts: TControl): Boolean;
begin
  {$IFDEF USERVKSDEVTE}
  Result := (ts as TTeTabSheet).PageVisible;
  {$ELSE}
  Result := (ts as TTabSheet).TabVisible;
  {$ENDIF}
end;

procedure TfrmRVBase.SetFormCaption(const Caption: String);
begin
  {$IFDEF USERVKSDEVTE}
  tef.Caption := _GetWideString(Caption);
  {$ELSE}
  Self.Caption := Caption;
  {$ENDIF}
end;

type
  TControlHack = class (TControl)
    public
      property Caption;
  end;

procedure TfrmRVBase.SetControlCaption(Control: TControl;
  const Caption: String);
begin
  {$IFDEF USERVKSDEVTE}
  if Control is TTeGroupBox then
    TTeGroupBox(Control).Caption := _GetWideString(Caption)
  else if Control is TTeButton then
    TTeButton(Control).Caption := _GetWideString(Caption)
  else  if Control is TTeRadioButton then
    TTeRadioButton(Control).Caption := _GetWideString(Caption)
  else if Control is TTeCheckbox then
    TTeCheckbox(Control).Caption := _GetWideString(Caption)
  else if Control is TTeLabel then
    TTeLabel(Control).Caption := _GetWideString(Caption)
  else if Control is TTeEdit then
    TTeEdit(Control).Text := Caption;
  {$ELSE}
  if Control is TEdit then
    TEdit(Control).Text := Caption
  else
    TControlHack(Control).Caption := Caption;
  {$ENDIF}
end;


function TfrmRVBase.GetEditText(Edit: TControl): String;
begin
  {$IFDEF USERVKSDEVTE}
  Result := TTeEdit(Edit).Text;
  {$ELSE}
  Result := TEdit(Edit).Text;
  {$ENDIF}
end;

function TfrmRVBase.GetEditFont(Edit: TControl): TFont;
begin
  {$IFDEF USERVKSDEVTE}
  Result := TTeEdit(Edit).Font;
  {$ELSE}
  Result := TEdit(Edit).Font;
  {$ENDIF}
end;

function TfrmRVBase.GetEditSelLength(Edit: TControl): Integer;
begin
  {$IFDEF USERVKSDEVTE}
  Result := TTeEdit(Edit).SelLength;
  {$ELSE}
  Result := TEdit(Edit).SelLength;
  {$ENDIF}
end;

function TfrmRVBase.GetEditSelStart(Edit: TControl): Integer;
begin
  {$IFDEF USERVKSDEVTE}
  Result := TTeEdit(Edit).SelStart;
  {$ELSE}
  Result := TEdit(Edit).SelStart;
  {$ENDIF}
end;

procedure TfrmRVBase.SetEditSelLength(Edit: TControl; Value: Integer);
begin
  {$IFDEF USERVKSDEVTE}
  TTeEdit(Edit).SelLength := Value;
  {$ELSE}
  TEdit(Edit).SelLength := Value;
  {$ENDIF}
end;

procedure TfrmRVBase.SetEditSelStart(Edit: TControl; Value: Integer);
begin
  {$IFDEF USERVKSDEVTE}
  TTeEdit(Edit).SelStart := Value;
  {$ELSE}
  TEdit(Edit).SelStart := Value;
  {$ENDIF}
end;

procedure TfrmRVBase.SetEditSelText(Edit: TControl; const s: String);
begin
  {$IFDEF USERVKSDEVTE}
  TTeEdit(Edit).InsertText(s);
  {$ELSE}
  TEdit(Edit).SelText := s;
  {$ENDIF}
end;

{$IFDEF USERVKSDEVTE}
function TfrmRVBase.IsTeFormAllowed: Boolean;
begin
  Result := True;
end;

function CompareTabOrders(Item1, Item2: Pointer): Integer;
begin
  if TControl(Item1) is TWinControl then
    if TControl(Item2) is TWinControl then
      Result := TWinControl(Item1).TabOrder-TWinControl(Item2).TabOrder
    else
      Result := 1
  else
    if TControl(Item2) is TWinControl then
      Result := -1
    else
      Result := 0;
end;


procedure TfrmRVBase.MoveChildren(OldControl, NewControl: TWinControl);
var i: Integer;
    Children: TList;
begin
  Children := TList.Create;
  for i := OldControl.ControlCount-1 downto 0 do
    Children.Add(OldControl.Controls[i]);
  Children.Sort(CompareTabOrders);
  for i := 0 to Children.Count-1 do
    TControl(Children[i]).Parent := NewControl;
end;

procedure TfrmRVBase.ApplyKSDevThemeEngine;
var teb: TTeButton;
    terb: TTeRadioButton;
    tecb: TTeCheckBox;
    tegb: TTeGroupBox;
    tecmb: TTeComboBox;
    tetxt: TTeEdit;    
    tervpp: TteRVPrintPreview;
    terv: TteRichView;
    tepc: TTePageControl;
    tets: TTeTabSheet;
    telbl: TTeLabel;
    terg: TTeRadioGroup;
    telst: TTeListBox;
    i, j: Integer;
    DialogBorder: Boolean;
    ATabOrder: Integer;
begin
   if IsTeFormAllowed then begin
     DialogBorder := BorderStyle=bsDialog;
     tef := TTeForm.Create(Self);
     if DialogBorder then begin
       BorderStyle := bsSizeable;
       tef.BorderStyle := kbsDialog;
       tef.Dragging.Sizeable := False;
     end;
     tef.BorderIcons := tef.BorderIcons - [kbiMaximize, kbiMinimize, kbiRollUp];
     RVA_Events.DoOnCreateTeForm(Self, tef);
     tef.Loaded;
   end;
   for i := ComponentCount-1 downto 0 do begin
     if not IsThemeAllowedFor(Components[i]) then
       continue;
     if Components[i] is TButton then
       with TButton(Components[i]) do begin
         teb := TTeButton.Create(Self);
         teb.SetBounds(Left,Top,Width,Height);
         teb.Caption := _GetWideString(Caption);
         teb.OnClick := OnClick;
         teb.ModalResult := ModalResult;
         if Self.Components[i] is TBitBtn then
          teb.Glyph := TBitBtn(Self.Components[i]).Glyph;
         teb.Hint := Hint;
         teb.ShowHint := ShowHint;
         teb.Enabled := Enabled;
         ATabOrder := TabOrder;
         teb.Default := Default;
         teb.Cancel := Cancel;
         teb.Parent := Parent;
         OnCreateThemedControl(TControl(Self.Components[i]), teb);
         Free;
         teb.Loaded;
         teb.TabOrder := ATabOrder;
       end
     else if Components[i].ClassName= 'TRadioButton' then
       with TRadioButton(Components[i]) do begin
         terb := TTeRadioButton.Create(Self);
         terb.SetBounds(Left,Top,Width,Height);
         terb.Caption := _GetWideString(Caption);
         terb.Enabled := Enabled;
         terb.Visible := Visible;
         terb.Checked := Checked;
         terb.OnClick := OnClick;
         terb.Hint := Hint;
         terb.ShowHint := ShowHint;
         ATabOrder := TabOrder;
         terb.Parent := Parent;
         OnCreateThemedControl(TControl(Self.Components[i]), terb);
         Free;
         terb.Loaded;
         terb.TabOrder := ATabOrder;
       end
     else if Components[i] is TCheckBox then
       with TCheckBox(Components[i]) do begin
         tecb := TTeCheckBox.Create(Self);
         tecb.SetBounds(Left,Top,Width,Height);
         tecb.Caption := _GetWideString(Caption);
         tecb.Enabled := Enabled;
         tecb.Visible := Visible;
         tecb.State := TTeCheckBoxState(State);
         tecb.OnClick := OnClick;
         tecb.Hint := Hint;
         tecb.ShowHint := ShowHint;
         ATabOrder := TabOrder;
         tecb.Parent := Parent;
         OnCreateThemedControl(TControl(Self.Components[i]), tecb);
         Free;
         tecb.Loaded;
         tecb.TabOrder := ATabOrder;
       end
     else if Components[i].ClassName='TEdit' then
       with TEdit(Components[i]) do begin
         tetxt := TTeEdit.Create(Self);
         tetxt.SetBounds(Left,Top,Width,Height);
         tetxt.Text := _GetWideString(Text);
         tetxt.Enabled := Enabled;
         tetxt.Visible := Visible;
         tetxt.OnChange := OnChange;
         tetxt.OnExit := OnExit;
         tetxt.OnKeyPress := OnKeyPress;
         tetxt.Hint := Hint;
         tetxt.ShowHint := ShowHint;
         ATabOrder := TabOrder;
         tetxt.Parent := Parent;
         OnCreateThemedControl(TControl(Self.Components[i]), tetxt);
         Free;
         tetxt.Loaded;
         tetxt.TabOrder := ATabOrder;
       end
     else if (Components[i].ClassName='TComboBox') and (TComboBox(Components[i]).Style<>csSimple) then
       with TComboBox(Components[i]) do begin
         tecmb := TTeComboBox.Create(Self);
         tecmb.SetBounds(Left,Top,Width,Height);
         tecmb.Enabled := Enabled;
         tecmb.Visible := Visible;
         tecmb.Items := Items;
         tecmb.ItemIndex := ItemIndex;
         tecmb.Hint := Hint;
         tecmb.ShowHint := ShowHint;
         ATabOrder := TabOrder;
         tecmb.Parent := Parent;
         tecmb.DropDownCount := DropDownCount;
         case Style of
           csDropDown:
             tecmb.ComboStyle := kcsDropDown;
           csDropDownList:
             tecmb.ComboStyle := kcsDropDownList;
           csOwnerDrawFixed:
             begin
               tecmb.ComboStyle := kcsDropDownList;
               tecmb.ListStyle := lbOwnerDrawFixed;
               tecmb.ItemHeight := ItemHeight;
               tecmb.Height := ItemHeight+2;
             end;
         end;
         tecmb.OnChange := OnClick;
         OnCreateThemedControl(TControl(Self.Components[i]), tecmb);
         Free;
         tecmb.Loaded;
         tecmb.TabOrder := ATabOrder;
       end
     else if (Components[i] is TGroupBox) then
       with TGroupBox(Components[i]) do begin
         tegb := TTeGroupBox.Create(Self);
         tegb.SetBounds(Left,Top,Width,Height);
         tegb.Caption := _GetWideString(Caption);
         tegb.Enabled := Enabled;
         tegb.Visible := Visible;
         ATabOrder := TabOrder;
         tegb.Parent := Parent;
         MoveChildren(TWinControl(Self.Components[i]), tegb);
         OnCreateThemedControl(TControl(Self.Components[i]), tegb);
         Free;
         tegb.Loaded;
         tegb.TabOrder := ATabOrder;
       end
     else if (Components[i] is TRadioGroup) then
       with TRadioGroup(Components[i]) do begin
         terg := TTeRadioGroup.Create(Self);
         terg.SetBounds(Left,Top,Width,Height);
         terg.Caption := _GetWideString(Caption);
         terg.Enabled := Enabled;
         terg.Visible := Visible;
         ATabOrder := TabOrder;
         terg.Parent := Parent;
         terg.Columns := Columns;
         for j := 0 to Items.Count-1 do
           terg.Items.Add({_GetWideString}(Items[j]));
         terg.ItemIndex := ItemIndex;
         terg.OnClick := OnClick;
         OnCreateThemedControl(TControl(Self.Components[i]), terg);
         Free;
         terg.Loaded;
         terg.TabOrder := ATabOrder;
       end
     else if (Components[i] is TListBox) then
       with TListBox(Components[i]) do begin
         telst := TTeListBox.Create(Self);
         telst.SetBounds(Left,Top,Width,Height);
         telst.Enabled := Enabled;
         telst.Visible := Visible;
         ATabOrder := TabOrder;
         telst.Parent := Parent;
         for j := 0 to Items.Count-1 do
           telst.Items.Add({_GetWideString}(Items[j]));
         telst.ItemIndex := ItemIndex;
         telst.OnClick := OnClick;
         telst.OnDblClick := OnDblClick;         
         OnCreateThemedControl(TControl(Self.Components[i]), telst);
         Free;
         telst.Loaded;
         telst.TabOrder := ATabOrder;
       end
     else if Components[i] is TLabel then
       with TLabel(Components[i]) do begin
         telbl := TTeLabel.Create(Self);
         telbl.SetBounds(Left,Top,Width,Height);
         telbl.Caption := _GetWideString(Caption);
         telbl.Enabled := Enabled;
         telbl.Visible := Visible;
         telbl.FocusControl := FocusControl;
         telbl.Parent := Parent;
         OnCreateThemedControl(TControl(Self.Components[i]), telbl);
         telbl.Loaded;
         Free;
       end
     else if Components[i] is TRVPrintPreview then
       with TRVPrintPreview(Components[i]) do begin
         tervpp := TteRVPrintPreview.Create(Self);
         tervpp.SetBounds(Left,Top,Width,Height);
         tervpp.OnZoomChanged := OnZoomChanged;
         ATabOrder := TabOrder;
         tervpp.Align := Align;
         tervpp.Parent := Parent;
         OnCreateThemedControl(TControl(Self.Components[i]), tervpp);
         Free;
         tervpp.TabOrder := ATabOrder;
       end
     else if (Components[i] is TRichView) and TRichView(Components[i]).VScrollVisible then
       with TRichView(Components[i]) do begin
         terv := TteRichView.Create(Self);
         terv.TabStop := TabStop;
         terv.SetBounds(Left,Top,Width,Height);
         terv.Style := Style;
         terv.AssignEvents(TRichView(Self.Components[i]));
         terv.Options := Options;
         ATabOrder := TabOrder;
         terv.Parent := Parent;
         OnCreateThemedControl(TControl(Self.Components[i]), terv);
         Free;
         terv.TabOrder := ATabOrder;
       end
     else if Components[i] is TPageControl then
       with TPageControl(Components[i]) do begin
         tepc := TTePageControl.Create(Self);
         tepc.SetBounds(Left,Top,Width,Height);
         ATabOrder := TabOrder;
         tepc.Parent := Parent;
         for j := 0 to PageCount-1 do begin
           tets := TTeTabSheet.Create(tepc);
           tepc.AddPage(tets);
           tets.PageIndex := Pages[j].PageIndex;
           tets.Caption := Pages[j].Caption;
           MoveChildren(Pages[j], tets);
         end;
         tepc.OnChange := OnChange;
         OnCreateThemedControl(TControl(Self.Components[i]), tepc);
         Free;
         tepc.Loaded;
         tepc.TabOrder := ATabOrder;
       end;

   end;
end;

procedure TfrmRVBase.OnCreateThemedControl(OldControl,
  NewControl: TControl);
begin

end;

function TfrmRVBase.IsThemeAllowedFor(Component: TComponent): Boolean;
begin
  Result := True;
end;
{$ENDIF}

end.
