
{*******************************************************}
{                                                       }
{       RichViewActions                                 }
{       Font attributes dialog                          }
{                                                       }
{       Copyright (c) 2002-2003, Sergey Tkachenko       }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

unit FontRVFrm;

interface

{$I RichViewActions.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, BaseRVFrm, RVScroll, RichView, StdCtrls, ComCtrls, ImgList,
  RVFontCombos, RVStyle, Buttons, RVTable, ColorRVFrm, RVColorGrid,
  RVSpinEdit, RVOfficeRadioBtn, RVColorCombo, RVALocalize;

type
  TfrmRVFont = class(TfrmRVBase)
    btnOk: TButton;
    btnCancel: TButton;
    ImageList1: TImageList;
    pc: TPageControl;
    ts1: TTabSheet;
    gbSample: TGroupBox;
    rv: TRichView;
    cmbSize: TRVFontSizeComboBox;
    cmbFont: TRVFontComboBox;
    cmbCharset: TRVFontCharsetComboBox;
    gbStyle: TGroupBox;
    cbB: TCheckBox;
    cbI: TCheckBox;
    gbEffects: TGroupBox;
    cbU: TCheckBox;
    cbS: TCheckBox;
    cbO: TCheckBox;
    cbAC: TCheckBox;
    lblFont: TLabel;
    lblSize: TLabel;
    lblScript: TLabel;
    rvs: TRVStyle;
    ts2: TTabSheet;
    gbSpacing: TGroupBox;
    lblSpacing: TLabel;
    rbCond: TRVOfficeRadioButton;
    rbExp: TRVOfficeRadioButton;
    seSpacing: TRVSpinEdit;
    gbVShift: TGroupBox;
    lblVShift: TLabel;
    rbUp: TRVOfficeRadioButton;
    rbDown: TRVOfficeRadioButton;
    seShift: TRVSpinEdit;
    Label5: TLabel;
    gbCharScale: TGroupBox;
    lblCharScale: TLabel;
    seCharScale: TRVSpinEdit;
    Label8: TLabel;
    cmbColor: TRVColorCombo;
    lblColor: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure pcChange(Sender: TObject);
    procedure cmbFontClick(Sender: TObject);
    procedure cmbCharsetClick(Sender: TObject);
    procedure rvsDrawStyleText(Sender: TRVStyle; const s: String;
      Canvas: TCanvas; StyleNo, SpaceBefore, Left, Top, Width,
      Height: Integer; DrawState: TRVTextDrawStates;
      var DoDefault: Boolean);
    procedure FormActivate(Sender: TObject);
    procedure rbExpClick(Sender: TObject);
    procedure seSpacingChange(Sender: TObject);
    procedure seShiftChange(Sender: TObject);
    procedure rbUpClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    table: TRVTableItemInfo;
    LockUpdates: Boolean;
    procedure UpdatePreview;
    function GetPreviewString: String;
  protected
    _gbSample, _pc: TControl;
    {$IFDEF USERVKSDEVTE}
    procedure OnCreateThemedControl(OldControl, NewControl: TControl); override;
    {$ENDIF}
  public
    { Public declarations }
    _cbB, _cbI, _cbU, _cbS, _cbO, _cbAC: TControl;
    ColorDialog: TColorDialog;
    PreviewString: String;
    procedure Localize; override;
  end;

implementation

{$R *.dfm}
{------------------------------------------------------------------------------}
procedure TfrmRVFont.FormCreate(Sender: TObject);
begin
  _gbSample := gbSample;
  _cbB  := cbB;
  _cbI  := cbI;
  _cbU  := cbU;
  _cbS  := cbS;
  _cbO  := cbO;
  _cbAC := cbAC;
  _pc   := pc;
  inherited;
  table := TRVTableItemInfo.CreateEx(1,1, rv.RVData);
  table.Cells[0,0].VAlign := rvcMiddle;
  table.Cells[0,0].Clear;
  table.Cells[0,0].AddNL('>>>   ',1,0);
  table.Cells[0,0].AddNL(RVA_GetS(rvam_font_Sample),0,-1);
  table.Cells[0,0].AddNL('   <<<',1,-1);
  table.Cells[0,0].BestHeight := rv.Height-2;
  table.CellPadding := 0;
  table.BorderVSpacing := 0;
  table.BorderHSpacing := 0;
  table.CellHSpacing := 0;
  table.CellVSpacing := 0;
  rv.AddItem('', table);
  rv.Format;
end;
{------------------------------------------------------------------------------}
function TfrmRVFont.GetPreviewString: String;
begin
  if PreviewString<>'' then
    Result := PreviewString
  else if cmbFont.ItemIndex<>-1 then
    Result := cmbFont.Items[cmbFont.ItemIndex]
  else
    Result := RVA_GetS(rvam_font_SampleText);
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.pcChange(Sender: TObject);
begin
  inherited;
  _gbSample.Parent := GetPageControlActivePage(_pc);
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.UpdatePreview;

   function GetCB(Style: TFontStyle): TControl;
   begin
     case Style of
       fsBold:   Result := _cbB;
       fsItalic: Result := _cbI;
       fsUnderline: Result := _cbU;
       fsStrikeOut: Result := _cbS;
       else Result := nil;
     end;
   end;

var i: TFontStyle;
    FontInfo: TFontInfo;
begin
  if not Visible then
    exit;

  if (cmbSize.Text<>'') then
    if StrToIntDef(cmbSize.Text,10)<1 then
      cmbSize.Text := '1';
    if StrToIntDef(cmbSize.Text,10)>100 then
      cmbSize.Text := '100';

  FontInfo := rvs.TextStyles[0];
  table.Cells[0,0].Items[1] := GetPreviewString;
  if cmbFont.ItemIndex>=0 then
    FontInfo.FontName := cmbFont.Items[cmbFont.ItemIndex]
  else
    FontInfo.FontName := 'Arial';
  FontInfo.Size := StrToIntDef(cmbSize.Text, 10);
  if cmbCharset.ItemIndex>=0 then
    FontInfo.Charset := cmbCharset.Charsets[cmbCharset.ItemIndex]
  else
    FontInfo.Charset := DEFAULT_CHARSET;
  for i := Low(TFontStyle) to High(TFontStyle) do
    if GetCheckBoxState(GetCB(i))<>cbChecked then
      FontInfo.Style := FontInfo.Style-[i]
    else
      FontInfo.Style := FontInfo.Style+[i];
  if GetCheckBoxState(_cbO)<>cbChecked then
    FontInfo.StyleEx := FontInfo.StyleEx-[rvfsOverline]
  else
    FontInfo.StyleEx := FontInfo.StyleEx+[rvfsOverline];
  if GetCheckBoxState(_cbAC)<>cbChecked then
    FontInfo.StyleEx := FontInfo.StyleEx-[rvfsAllCaps]
  else
    FontInfo.StyleEx := FontInfo.StyleEx+[rvfsAllCaps];

  if rbDown.Checked then
    FontInfo.VShift := -seShift.AsInteger
  else
    FontInfo.VShift := seShift.AsInteger;

  if cmbColor.Indeterminate then
    FontInfo.Color := clWindowText
  else
    FontInfo.Color := cmbColor.ChosenColor;

  if rbCond.Checked then
    FontInfo.CharSpacing := -seSpacing.AsInteger
  else
    FontInfo.CharSpacing := seSpacing.AsInteger;

  if seCharScale.Indeterminate then
    FontInfo.CharScale := 100
  else
    FontInfo.CharScale := seCharScale.AsInteger;

  table.Changed;
  rv.Format;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.cmbFontClick(Sender: TObject);
begin
  if cmbFont.ItemIndex>=0 then begin
    LockUpdates := True;
    cmbSize.FontName :=  cmbFont.Items[cmbFont.ItemIndex];
    cmbCharset.FontName :=  cmbFont.Items[cmbFont.ItemIndex];
    LockUpdates := False;
  end;
  UpdatePreview;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.cmbCharsetClick(Sender: TObject);
begin
  UpdatePreview;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.rvsDrawStyleText(Sender: TRVStyle; const s: String;
  Canvas: TCanvas; StyleNo, SpaceBefore, Left, Top, Width, Height: Integer;
  DrawState: TRVTextDrawStates; var DoDefault: Boolean);
var tm: TTextMetric;
    bl: Integer;
begin
  if Sender.ItemNo<>0 then
    exit;
  if GetTextMetrics(Canvas.Handle, tm) then begin
    bl := tm.tmAscent+Top;
    Canvas.Pen.Color := Canvas.Font.Color;
    Canvas.Pen.Width := 1;
    Canvas.Pen.Style := psSolid;
    Canvas.MoveTo(0,bl);
    Canvas.LineTo(rv.ClientWidth+1,bl);
  end;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.FormActivate(Sender: TObject);
begin
  inherited;
  cmbSize.ItemIndex := cmbSize.Items.IndexOf(cmbSize.Text);
  cmbColor.ColorDialog := ColorDialog;
  seSpacingChange(Sender);
  seShiftChange(Sender);
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.rbExpClick(Sender: TObject);
begin
  if not Visible then
    exit;
  if seSpacing.Indeterminate and (rbExp.Checked or rbCond.Checked) then
    seSpacing.Value := 1;
  if rbExp.Checked then
    seSpacing.MaxValue := 100
  else begin
    seSpacing.MaxValue := 20;
    seSpacing.Value := seSpacing.Value;
  end;
  UpdatePreview;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.seSpacingChange(Sender: TObject);
begin
  if not Visible then
    exit;
  if not seSpacing.Indeterminate and not rbExp.Checked and not rbCond.Checked then begin
    rbExp.Checked := True;
    seSpacing.MaxValue := 100;
    end
  else if seSpacing.Indeterminate then begin
    rbExp.Checked := False;
    rbCond.Checked := False;
  end;
  UpdatePreview;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.rbUpClick(Sender: TObject);
begin
  if not Visible then
    exit;
  if seShift.Indeterminate and (rbUp.Checked or rbDown.Checked) then
    seShift.Value := 1;
  UpdatePreview;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.seShiftChange(Sender: TObject);
begin
  if not Visible then
    exit;
  if not seShift.Indeterminate and not rbUp.Checked and not rbDown.Checked then
    rbDown.Checked := True
  else if seShift.Indeterminate then begin
    rbUp.Checked := False;
    rbDown.Checked := False;
  end;
  UpdatePreview;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  // Strange... combo boxes eat Esc
  if Key=VK_ESCAPE then
    ModalResult := mrCancel;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVFont.Localize;
begin
  inherited;
  Caption := RVA_GetS(rvam_font_Title);
  btnOk.Caption := RVA_GetS(rvam_btn_OK);
  btnCancel.Caption := RVA_GetS(rvam_btn_Cancel);  
  ts1.Caption := RVA_GetSH(rvam_font_FontTab);
  ts2.Caption := RVA_GetSH(rvam_font_LayoutTab);
  lblFont.Caption := RVA_GetS(rvam_font_FontName);
  lblSize.Caption := RVA_GetS(rvam_font_FontSize);
  gbStyle.Caption := RVA_GetSH(rvam_font_FontStyle);
  cbB.Caption := RVA_GetS(rvam_font_Bold);
  cbI.Caption := RVA_GetS(rvam_font_Italic);
  lblScript.Caption := RVA_GetS(rvam_font_Script);
  lblColor.Caption := RVA_GetS(rvam_font_Color);
  cmbCharset.DefaultCharsetCaption := RVA_GetS(rvam_font_DefaultCharset);
  gbEffects.Caption := RVA_GetSH(rvam_font_Effects);
  cbU.Caption := RVA_GetS(rvam_font_Underline);
  cbO.Caption := RVA_GetS(rvam_font_Overline);
  cbS.Caption := RVA_GetS(rvam_font_Strikethrough);
  cbAC.Caption := RVA_GetS(rvam_font_AllCaps);

  gbSample.Caption  := RVA_GetSH(rvam_font_Sample);
  gbSpacing.Caption := RVA_GetSH(rvam_font_SpacingH);
  lblSpacing.Caption := RVA_GetS(rvam_font_Spacing);
  rbExp.Caption     := RVA_GetS(rvam_font_Expanded);
  rbCond.Caption    := RVA_GetS(rvam_font_Condensed);

  gbVShift.Caption  := RVA_GetSH(rvam_font_OffsetH);
  lblVShift.Caption := RVA_GetS(rvam_font_Offset);
  rbDown.Caption    := RVA_GetS(rvam_font_Down);
  rbUp.Caption      := RVA_GetS(rvam_font_Up);

  gbCharScale.Caption := RVA_GetSH(rvam_font_ScalingH);
  lblCharScale.Caption := RVA_GetS(rvam_font_Scaling);  
end;

{$IFDEF USERVKSDEVTE}
procedure TfrmRVFont.OnCreateThemedControl(OldControl,
  NewControl: TControl);
begin
  if OldControl = _gbSample then
    _gbSample := NewControl
  else if OldControl = _cbB then
    _cbB := NewControl
  else if OldControl = _cbI then
    _cbI := NewControl
  else if OldControl = _cbS then
    _cbS := NewControl
  else if OldControl = _cbO then
    _cbO := NewControl
  else if OldControl = _cbU then
    _cbU := NewControl
  else if OldControl = _cbAC then
    _cbAC := NewControl
  else if OldControl = _pc then
    _pc := NewControl;
end;
{$ENDIF}

end.
