
{*******************************************************}
{                                                       }
{       RichViewActions                                 }
{       Insert-symbol dialog                            }
{                                                       }
{       Copyright (c) 2002-2003, Sergey Tkachenko       }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

unit InsSymbolRVFrm;

interface

{$I RV_Defs.inc}
{$I RichViewActions.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, BaseRVFrm, RVFontCombos, StdCtrls, ExtCtrls, RVALocalize,
  RVGrids;

const
  CM_DENYSUBCLASSING = CM_BASE + 2000;

type

  TZoomPanel = class (TPanel)
  private
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure CMDenySubclassing(var Msg: TMessage); message CM_DENYSUBCLASSING;
  protected
    procedure Paint; override;
  public
    DefWidth: Integer;
    FontName: String;
    FontCharset: TFontCharset;
    Text: String;
    TextW: WideString;
  end;

  TfrmRVInsertSymbol = class(TfrmRVBase)
    btnOk: TButton;
    btnCancel: TButton;
    lblCharset: TLabel;
    lblFont: TLabel;
    cmbFont: TRVFontComboBox;
    cmbCharset: TRVFontCharsetComboBox;
    dg: TRVGrid;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure dgDrawCell(Sender: TObject; ACol, ARow: Integer; ARect: TRect;
      Selected: Boolean);
    procedure cmbFontClick(Sender: TObject);
    procedure cmbCharsetClick(Sender: TObject);
    procedure dgSelectCell(Sender: TObject);
    procedure dgTopLeftChanged(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure dgEnter(Sender: TObject);
  private
    { Private declarations }
    FontName: String;
    FontCharset: TFontCharset;
  protected
    {$IFDEF USERVKSDEVTE}
    procedure OnCreateThemedControl(OldControl, NewControl: TControl); override;
    {$ENDIF}
  public
    { Public declarations }
    Panel: TZoomPanel;
    _btnOk, _lblCharset: TControl;
    procedure Init(Char: Word; const AFontName: String; ACharset: TFontCharset);
    procedure GetInfo(var Char: Word; var AFontName: String; var ACharset: TFontCharset);
    procedure SetOptions(AllowUnicode, AllowANSI: Boolean);
    procedure Localize; override;
  end;


implementation

uses math; // for Min()

{$R *.dfm}

procedure TfrmRVInsertSymbol.FormCreate(Sender: TObject);
begin
  _btnOk := btnOk;
  _lblCharset := lblCharset;
  inherited;
  // defaults are not needed
  //dg.DefaultColWidth := (dg.Width-4-GetSystemMetrics(SM_CXVSCROLL)) div (dg.ColCount+1);
  //dg.DefaultRowHeight := (dg.ClientHeight) div (dg.RowCount+1);
  //dg.Height := dg.RowCount*(dg.DefaultRowHeight+1)+(dg.Height-dg.ClientHeight);

  Panel := TZoomPanel.Create(Self);
  Panel.Parent := Self;
  Panel.Visible := False;

  Init(183, 'Symbol', SYMBOL_CHARSET);
end;

procedure TfrmRVInsertSymbol.Init(Char: Word; const AFontName: String;
  ACharset: TFontCharset);
var idx: Integer;
begin
  cmbFont.ItemIndex := cmbFont.Items.IndexOf(AFontName);
  if cmbFont.ItemIndex<0 then
    cmbFont.ItemIndex := 0;
  cmbFontClick(nil);
  idx := cmbCharset.IndexOfCharset(ACharset);
  if idx>=0 then
    cmbCharset.ItemIndex := idx;
  cmbCharsetClick(nil);
  if Char=0 then
    Char := 32;
  dec(Char, 32);
  dg.SelectCell(Char mod 28, Char div 28);
end;

procedure TfrmRVInsertSymbol.SetOptions(AllowUnicode, AllowANSI: Boolean);
var idx: Integer;
begin
  if not AllowANSI then begin
    idx := cmbCharset.IndexOfCharset(DEFAULT_CHARSET);
    if idx>=0 then
      cmbCharset.ItemIndex := idx;
    cmbCharset.Visible := False;
    _lblCharset.Visible := False;
  end;
  if not AllowUnicode then begin
    cmbCharset.AddDefaultCharset := False;
  end;
  cmbCharsetClick(nil);
end;

procedure TfrmRVInsertSymbol.GetInfo(var Char: Word;
  var AFontName: String; var ACharset: TFontCharset);
begin
  Char := 32+dg.Row*dg.ColCount+dg.Col;
  AFontName := FontName;
  ACharset  := FontCharset;
end;

procedure TfrmRVInsertSymbol.dgDrawCell(Sender: TObject; ACol,
  ARow: Integer; ARect: TRect; Selected: Boolean);
var sz: TSize;
    s: String;
    ws: WideString;
begin
  dg.Canvas.FillRect(ARect);
  dg.Canvas.Font.Name := FontName;
  dg.Canvas.Font.Charset := FontCharset;
  dg.Canvas.Font.Style := [];
  dg.Canvas.Font.Size := MulDiv(Min(dg.Width, dg.Height * 2), 10, 320);
  if FontCharset<>DEFAULT_CHARSET then begin
    s := Char(32+ARow*dg.ColCount+ACol);
    sz := dg.Canvas.TextExtent(s);
    sz.cx := (ARect.Right-ARect.Left-sz.cx) div 2 + ARect.Left;
    sz.cy := (ARect.Bottom-ARect.Top-sz.cy) div 2 + ARect.Top;
    dg.Canvas.TextOut(sz.cx, sz.cy, s);
    end
  else begin
    if 32+ARow*dg.ColCount+ACol>$FFFF then
      exit;
    ws := WideChar(32+ARow*dg.ColCount+ACol);
    GetTextExtentPoint32W(dg.Canvas.Handle, PWideChar(ws), 1, sz);
    sz.cx := (ARect.Right-ARect.Left-sz.cx) div 2 + ARect.Left;
    sz.cy := (ARect.Bottom-ARect.Top-sz.cy) div 2 + ARect.Top;
    TextOutW(dg.Canvas.Handle, sz.cx, sz.cy, PWideChar(ws), 1);
  end;
end;

procedure TfrmRVInsertSymbol.cmbFontClick(Sender: TObject);
begin
  if cmbFont.ItemIndex<0 then
    exit;
  dg.repaint;
  FontName := cmbFont.Items[cmbFont.ItemIndex];
  if Panel<>nil then begin
    Panel.FontName := FontName;
    Panel.Refresh;
  end;
  cmbCharset.FontName := FontName;
end;

procedure TfrmRVInsertSymbol.cmbCharsetClick(Sender: TObject);
begin
  FontCharset := cmbCharset.Charsets[cmbCharset.ItemIndex];
  if Panel<>nil then
    Panel.FontCharset := FontCharset;
  if FontCharset=DEFAULT_CHARSET then
    dg.RowCount := 2340
  else
    dg.RowCount := 8;
  dgSelectCell(dg);
end;

procedure TfrmRVInsertSymbol.dgSelectCell(Sender: TObject);
var
  ARow, ACol: Integer;
  ARect: TRect;
begin
  _btnOk.Enabled := True;
  ARow := Max(dg.Row, 0);
  ACol := Max(dg.Col, 0);
  if FontCharset<>DEFAULT_CHARSET then
    Label1.Caption := Format(RVA_GetS(rvam_is_CharCode), [32+ARow*dg.ColCount+ACol])
  else if 32+ARow*dg.ColCount+ACol>$FFFF then begin
    Label1.Caption := RVA_GetS(rvam_is_NoChar);
    _btnOk.Enabled := False;
    end
  else
    Label1.Caption := Format(RVA_GetS(rvam_is_UCharCode), [32+ARow*dg.ColCount+ACol]);
  if Panel=nil then
    exit;
  if FontCharset<>DEFAULT_CHARSET then
    Panel.Text := Chr(32+ARow*dg.ColCount+ACol)
  else begin
    if 32+ARow*dg.ColCount+ACol>$FFFF then
      Panel.TextW := ''
    else
      Panel.TextW := WideChar(32+ARow*dg.ColCount+ACol);
  end;
  // The code below replaces the huge .SetBounds that was here before
  ARect := dg.GetCellRect(ACol, ARow);
  InflateRect(ARect, 10, 10);
  OffsetRect(ARect, dg.Left, dg.Top);
  Panel.DefWidth := ARect.Right - ARect.Left;
  with ARect do Panel.SetBounds(Left, Top, Right - Left, Bottom - Top);
  // ---
  Panel.Refresh;
end;

{ TZoomPanel }

procedure TZoomPanel.CMDenySubclassing(var Msg: TMessage);
begin
  Msg.Result := 1;
end;

procedure TZoomPanel.Paint;
var r: TRect;
    sz: TSize;
    w: Integer;
begin
  Canvas.Brush.Color := clHighlight;
  r := ClientRect;
  Canvas.FillRect(r);
  Canvas.Font.Height := Height-10;
  Canvas.Font.Name := FontName;
  Canvas.Font.Charset := FontCharset;
  Canvas.Font.Color := clHighlightText;
  Canvas.Pen.Color := clActiveBorder;
  Canvas.Rectangle(0,0,r.Right,r.Bottom);
  Canvas.Brush.Style := bsClear;
  if FontCharset<>DEFAULT_CHARSET then begin
    sz := Canvas.TextExtent(Text);
    w := sz.cx;
    sz.cx := (ClientWidth-sz.cx) div 2;
    sz.cy := (ClientHeight-sz.cy) div 2;
    Canvas.TextOut(sz.cx, sz.cy, Text);
    end
  else begin
    GetTextExtentPoint32W(Canvas.Handle, PWideChar(TextW), Length(TextW), sz);
    w := sz.cx;    
    sz.cx := (ClientWidth-sz.cx) div 2;
    sz.cy := (ClientHeight-sz.cy) div 2;
    TextOutW(Canvas.Handle, sz.cx, sz.cy, PWideChar(TextW), Length(TextW));
  end;
  if w > DefWidth then
    Width := w;
{  else
    Width := DefWidth;}
end;

procedure TZoomPanel.WMNCHitTest(var Message: TWMNCHitTest);
begin
  Message.Result := HTTRANSPARENT;
end;

procedure TfrmRVInsertSymbol.Localize;
begin
  inherited;
  Caption := RVA_GetS(rvam_is_Title);
  btnOk.Caption := RVA_GetS(rvam_btn_Insert);
  btnCancel.Caption := RVA_GetS(rvam_btn_Cancel);
  lblFont.Caption := RVA_GetS(rvam_is_Font);
  lblCharset.Caption := RVA_GetS(rvam_is_Charset);
  cmbCharset.DefaultCharsetCaption := RVA_GetS(rvam_is_Unicode);
end;

procedure TfrmRVInsertSymbol.dgTopLeftChanged(Sender: TObject);
begin
  dgSelectCell(dg); // actually, onSelectCell is called twice...
                    // see DoTopLeftChanged first two "if"s
                    // but it doesn't matter IMHO
end;

procedure TfrmRVInsertSymbol.FormResize(Sender: TObject);
begin
  dgSelectCell(dg); // to adjust Zoompanel position
end;

procedure TfrmRVInsertSymbol.dgEnter(Sender: TObject);
begin
  Panel.Visible := True; // also to adjust Zoompanel position
end;

{$IFDEF USERVKSDEVTE}
procedure TfrmRVInsertSymbol.OnCreateThemedControl(OldControl,
  NewControl: TControl);
begin
  if OldControl = _btnOk then
    _btnOk := NewControl
  else if OldControl = _lblCharset then
    _lblCharset := NewControl;
end;
{$ENDIF}

end.
