
{*******************************************************}
{                                                       }
{       RichViewActions                                 }
{       Print-preview dialog                            }
{                                                       }
{       Copyright (c) 2002-2003, Sergey Tkachenko       }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

unit PreviewRVFrm;

interface

{$I RV_Defs.inc}
{$I RichViewActions.inc}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, RVScroll, CRVPP, RVPP, ExtCtrls, BaseRVFrm, ActnList, ImgList,
  Buttons, ComCtrls, RVALocalize, RichViewActions, PtblRV;

type
  TfrmRVPreview = class(TfrmRVBase)
    Panel1: TPanel;
    rvpp: TRVPrintPreview;
    cmb1: TComboBox;
    Label1: TLabel;
    SpeedButton1: TSpeedButton;
    ActionList1: TActionList;
    ImageList1: TImageList;
    actFirst: TAction;
    actPrior: TAction;
    actNext: TAction;
    actLast: TAction;
    SpeedButton2: TSpeedButton;
    SpeedButton3: TSpeedButton;
    SpeedButton4: TSpeedButton;
    txtPageNo: TEdit;
    Label2: TLabel;
    Button1: TButton;
    SpeedButton5: TSpeedButton;
    actPrint: TAction;
    rvActionPageSetup1: TrvActionPageSetup;
    SpeedButton6: TSpeedButton;
    procedure cmb1Exit(Sender: TObject);
    procedure cmb1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure rvppZoomChanged(Sender: TObject);
    procedure actFirstExecute(Sender: TObject);
    procedure actPriorExecute(Sender: TObject);
    procedure actNextExecute(Sender: TObject);
    procedure actLastExecute(Sender: TObject);
    procedure txtPageNoKeyPress(Sender: TObject; var Key: Char);
    procedure txtPageNoExit(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure actPrintExecute(Sender: TObject);
    procedure rvActionPageSetup1Change(Sender: TObject);
  private
    { Private declarations }
    {$IFDEF RICHVIEWDEF6}
    cmb: TComboBoxEx;
    {$ELSE}
    cmb: TComboBox;
    {$ENDIF}
    procedure UpdateZoom;
    procedure UpdateToolbar;
    procedure GoToPage;
  protected
    _txtPageNo, _Label2: TControl;
    {$IFDEF USERVKSDEVTE}
    function IsThemeAllowedFor(Component: TComponent): Boolean; override;
    procedure OnCreateThemedControl(OldControl, NewControl: TControl); override;
    {$ENDIF}
  public
    { Public declarations }
    _rvpp: TRVPrintPreview;
    procedure Localize; override;
  end;


implementation

resourcestring

 sZoom200 = '200%';
 sZoom150 = '150%';
 sZoom100 = '100%';
 sZoom75  = '75%';
 sZoom50  = '50%';
 sZoom25  = '25%';
 sZoom10  = '10%';

{$R *.DFM}
{-----------------------------------------------------------------------}
procedure TfrmRVPreview.cmb1Exit(Sender: TObject);
begin
  UpdateZoom;
end;
{-----------------------------------------------------------------------}
procedure TfrmRVPreview.cmb1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_RETURN then begin
    UpdateZoom;
    Key := 0;
  end;
end;
{-----------------------------------------------------------------------}
procedure TfrmRVPreview.UpdateZoom;
var s: String;
    zoom: Integer;
begin
  s := Trim(cmb.Text);
  if s=RVA_GetS(rvam_pp_PageWidth) then begin
    _rvpp.ZoomMode := rvzmPageWidth;
    exit;
  end;
  if s=RVA_GetS(rvam_pp_FullPage) then begin
    _rvpp.ZoomMode := rvzmFullPage;
    exit;
  end;
  if (s<>'') and (s[Length(s)]='%') then
    s := Copy(s,1,Length(s)-1);
  zoom := StrToIntDef(s,0);
  if (zoom<10) or (zoom>500) then
    Application.MessageBox(RVA_GetPC(rvam_pp_InvalidScale),RVA_GetPC(rvam_pp_Title), MB_OK or MB_ICONEXCLAMATION)
  else
    _rvpp.SetZoom(zoom);
end;
{-----------------------------------------------------------------------}
procedure TfrmRVPreview.rvppZoomChanged(Sender: TObject);
begin
  cmb.Text := IntToStr(_rvpp.ZoomPercent)+'%';
end;
{-----------------------------------------------------------------------}
procedure TfrmRVPreview.UpdateToolbar;
var delta: Integer;
begin
  actFirst.Enabled := _rvpp.PageNo>1;
  actPrior.Enabled := _rvpp.PageNo>1;
  actNext.Enabled := _rvpp.PageNo<_rvpp.RVPrint.PagesCount;
  actLast.Enabled := _rvpp.PageNo<_rvpp.RVPrint.PagesCount;
  SetControlCaption(_Label2, Format(RVA_GetS(rvam_pp_OfNo), [_rvpp.RVPrint.PagesCount]));
  SetControlCaption(_txtPageNo, IntToStr(_rvpp.PageNo));
  delta := _Label2.Left+_Label2.Width+5-SpeedButton3.Left;
  SpeedButton3.Left := SpeedButton3.Left+Delta;
  SpeedButton4.Left := SpeedButton4.Left+Delta;
end;
{-----------------------------------------------------------------------}
procedure TfrmRVPreview.actFirstExecute(Sender: TObject);
begin
  _rvpp.First;
  UpdateToolbar;
end;

procedure TfrmRVPreview.actPriorExecute(Sender: TObject);
begin
  _rvpp.Prev;
  UpdateToolbar;
end;

procedure TfrmRVPreview.actNextExecute(Sender: TObject);
begin
  _rvpp.Next;
  UpdateToolbar;
end;

procedure TfrmRVPreview.actLastExecute(Sender: TObject);
begin
  _rvpp.Last;
  UpdateToolbar;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVPreview.txtPageNoKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = #13) then begin
    Key := #0;
    GoToPage;
  end;
  if not (Key in ['0'..'9']) and (Key >= #32) then begin
    Beep;
    Key := #0;
  end;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVPreview.txtPageNoExit(Sender: TObject);
begin
  GoToPage
end;
{------------------------------------------------------------------------------}
procedure TfrmRVPreview.GoToPage;
var pn: Integer;
begin
  pn := StrToIntDef(GetEditText(_txtPageNo), _rvpp.PageNo);
  if pn<1 then
    pn := 1;
  if pn>_rvpp.RVPrint.PagesCount then
    pn := _rvpp.RVPrint.PagesCount;
  _rvpp.PageNo := pn;
  UpdateToolbar;
end;
{------------------------------------------------------------------------------}

procedure TfrmRVPreview.FormCreate(Sender: TObject);
begin
  _rvpp := rvpp;
  _txtPageNo := txtPageNo;
  _Label2 := Label2;
  inherited;
  Panel1.DoubleBuffered := True;
  {$IFDEF RICHVIEWDEF6}
  cmb := TComboBoxEx.Create(Self);
  cmb.Parent := Panel1;
  with cmb1 do
    cmb.SetBounds(Left,Top,Width,Height);
  cmb.TabOrder := 1;
  cmb.OnExit := cmb1.OnExit;
  cmb.OnClick := cmb1.OnClick;  
  cmb.OnKeyDown := cmb1.OnKeyDown;
  cmb1.Free;
  cmb.Images := ImageList1;
  with cmb.ItemsEx do begin
   with Add do
     Caption := sZoom200;
   with Add do
     Caption := sZoom150;
   with Add do
     Caption := sZoom100;
   with Add do
     Caption := sZoom75;
   with Add do
     Caption := sZoom50;
   with Add do
     Caption := sZoom25;
   with Add do
     Caption := sZoom10;
   with Add do begin
     Caption := RVA_GetS(rvam_pp_PageWidth);
     ImageIndex := 4;
   end;
   with Add do begin
     Caption := RVA_GetS(rvam_pp_FullPage);
     ImageIndex := 5;
   end;
  end;
  {$ELSE}
  cmb := cmb1;
  with cmb.Items do begin
   Add(sZoom200);
   Add(sZoom150);
   Add(sZoom100);
   Add(sZoom75);
   Add(sZoom50);
   Add(sZoom25);
   Add(sZoom10);
   Add(RVA_GetS(rvam_pp_PageWidth));
   Add(RVA_GetS(rvam_pp_FullPage));
  end;
  {$ENDIF}
  cmb.DropDownCount := 10;
end;
{------------------------------------------------------------------------------}
procedure TfrmRVPreview.actPrintExecute(Sender: TObject);
begin
  ModalResult := mrOk;
end;

procedure TfrmRVPreview.Localize;
begin
  inherited;
  Caption := RVA_GetS(rvam_pp_Title);
  Button1.Caption := RVA_GetS(rvam_btn_Close);
  Label1.Caption := RVA_GetS(rvam_pp_Pages);
  SpeedButton1.Hint := RVA_GetS(rvam_pp_First);
  SpeedButton2.Hint := RVA_GetS(rvam_pp_Prior);
  SpeedButton3.Hint := RVA_GetS(rvam_pp_Next);
  SpeedButton4.Hint := RVA_GetS(rvam_pp_Last);
  actPrint.Hint := RVA_GetS(rvam_act_PrintH);
  RVA_LocalizeForm(Self);
  SpeedButton6.Caption := '';
end;

procedure TfrmRVPreview.rvActionPageSetup1Change(Sender: TObject);
begin
  TRVPrint(_rvpp.RVPrint).FormatPages(rvdoALL);
  _rvpp.UpdateView;
end;

{$IFDEF USERVKSDEVTE}
function TfrmRVPreview.IsThemeAllowedFor(Component: TComponent): Boolean;
begin
  Result := Component<>cmb1;

end;

procedure TfrmRVPreview.OnCreateThemedControl(OldControl,
  NewControl: TControl);
begin
  if OldControl = _rvpp then
    _rvpp := TRVPrintPreview(NewControl)
  else if OldControl = _txtPageNo then
    _txtPageNo := NewControl
  else if OldControl = _Label2 then
    _Label2 := NewControl;
end;

{$ENDIF}



end.
